/**  \file  suite_oomstub.c
 *  \brief  Unit testing for malloc/free/realloc stubs
 * \author  François Cerbelle (Fanfan), francois@cerbelle.net
 *
 * \internal
 *       Created:  07/07/2024
 *      Revision:  none
 * Last modified:  2024-12-11 15:55
 *      Compiler:  gcc
 *  Organization:  Cerbelle.net
 *     Copyright:  Copyright (c) 2024, François Cerbelle
 *
 *  This source code is released for free distribution under the terms of the
 *  GNU General Public License as published by the Free Software Foundation.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "liboom/oomstub.h"                        /* OOM simulation */
#include "checktools.inc"
#include <check.h>
#include <stdlib.h>                             /* abort() */

START_TEST(oomstub_malloc) {
    char* a=NULL;
    int d1,d2;

    oomstub_setcountdown(0);
    d1 = oomstub_getcountdown();
    a = malloc(1);
    d2 = oomstub_getcountdown();
    oomstub_setcountdown(0);
    ck_assert(0==d1);
    ck_assert(0==d2);
    ck_assert(NULL!=a);
    free(a);

    oomstub_setcountdown(1);
    d1 = oomstub_getcountdown();
    a = malloc(1);
    d2 = oomstub_getcountdown();
    oomstub_setcountdown(0);
    ck_assert(1==d1);
    ck_assert(0==d2);
    ck_assert(NULL==a);

    oomstub_setcountdown(2);
    d1 = oomstub_getcountdown();
    a = malloc(1); /* cppcheck-suppress memleak */
    d2 = oomstub_getcountdown();
    oomstub_setcountdown(0);
    ck_assert(2==d1);
    ck_assert(1==d2);
    ck_assert(NULL!=a);
    free(a);
}
END_TEST

START_TEST(oomstub_calloc) {
    char* a=NULL;
    int d1,d2;

    oomstub_setcountdown(0);
    d1 = oomstub_getcountdown();
    a = calloc(5,2);
    d2 = oomstub_getcountdown();
    oomstub_setcountdown(0);
    ck_assert(0==d1);
    ck_assert(0==d2);
    ck_assert(NULL!=a);
    free(a);

    oomstub_setcountdown(1);
    d1 = oomstub_getcountdown();
    a = calloc(5,2);
    d2 = oomstub_getcountdown();
    oomstub_setcountdown(0);
    ck_assert(1==d1);
    ck_assert(0==d2);
    ck_assert(NULL==a);

    oomstub_setcountdown(2);
    d1 = oomstub_getcountdown();
    a = calloc(5,2); /* cppcheck-suppress memleak */
    d2 = oomstub_getcountdown();
    oomstub_setcountdown(0);
    ck_assert(2==d1);
    ck_assert(1==d2);
    ck_assert(NULL!=a);
    free(a);
}
END_TEST

/* Using NULL as a pointer to reallocate means malloc behavior */
START_TEST(oomstub_realloc_malloc) {
    void* newptr;
    int d1,d2;

    oomstub_setcountdown(0); /* No failure */
    d1 = oomstub_getcountdown();
    newptr = realloc(NULL,1);
    d2 = oomstub_getcountdown();
    oomstub_setcountdown(0);
    ck_assert(0==d1);
    ck_assert(0==d2);
    ck_assert(NULL!=newptr);
    free(newptr);

    oomstub_setcountdown(1); /* Immediate failure */
    d1 = oomstub_getcountdown();
    newptr = realloc(NULL,1);
    d2 = oomstub_getcountdown();
    oomstub_setcountdown(0);
    ck_assert(1==d1);
    ck_assert(0==d2);
    ck_assert(NULL==newptr);

    oomstub_setcountdown(2); /* Success first before failure */
    d1 = oomstub_getcountdown();
    newptr = realloc(NULL,1); /* cppcheck-suppress memleak */
    d2 = oomstub_getcountdown();
    oomstub_setcountdown(0);
    ck_assert(2==d1);
    ck_assert(1==d2);
    ck_assert(NULL!=newptr);
    free(newptr);
}
END_TEST

/* Using 0 as a size to reallocate means free behavior */
START_TEST(oomstub_realloc_free) {
    void* oldptr;
    void* newptr;
    int d1,d2;

    oldptr = malloc(1);
    oomstub_setcountdown(0); /* No failure */
    d1 = oomstub_getcountdown();
    newptr = realloc(oldptr,0);
    d2 = oomstub_getcountdown();
    oomstub_setcountdown(0);
    ck_assert(0==d1);
    ck_assert(0==d2);
    ck_assert(NULL==newptr);
    free(newptr);

    oldptr = malloc(1);
    oomstub_setcountdown(1); /* Immediate failure */
    d1 = oomstub_getcountdown();
    newptr = realloc(oldptr,0);
    d2 = oomstub_getcountdown();
    oomstub_setcountdown(0);
    ck_assert(1==d1);
    ck_assert(0==d2);
    ck_assert(NULL!=newptr);
    free(newptr); /* Not actually freed because of simulated failure */

    oldptr = malloc(1);
    oomstub_setcountdown(2); /* Success first before failure */
    d1 = oomstub_getcountdown();
    newptr = realloc(oldptr,0);
    d2 = oomstub_getcountdown();
    oomstub_setcountdown(0);
    ck_assert(2==d1);
    ck_assert(1==d2);
    ck_assert(NULL==newptr);
    free(newptr);
}
END_TEST

START_TEST(oomstub_realloc_downsize) {
    void* oldptr;
    void* newptr;
    int d1,d2;

    oldptr = malloc(1000);
    oomstub_setcountdown(0); /* No failure */
    d1 = oomstub_getcountdown();
    newptr = realloc(oldptr,10);
    d2 = oomstub_getcountdown();
    oomstub_setcountdown(0);
    ck_assert(0==d1);
    ck_assert(0==d2);
    ck_assert(NULL!=newptr);
    free(newptr);

    oldptr = malloc(1000);
    oomstub_setcountdown(1); /* Immediate failure */
    d1 = oomstub_getcountdown();
    newptr = realloc(oldptr,10);
    d2 = oomstub_getcountdown();
    oomstub_setcountdown(0);
    ck_assert(1==d1);
    ck_assert(0==d2);
    ck_assert(NULL==newptr);
    free(oldptr); /* cppcheck-suppress doubleFree */

    oldptr = malloc(1000);
    oomstub_setcountdown(2); /* Success first before failure */
    d1 = oomstub_getcountdown();
    newptr = realloc(oldptr,10); /* cppcheck-suppress memleak */
    d2 = oomstub_getcountdown();
    oomstub_setcountdown(0);
    ck_assert(2==d1);
    ck_assert(1==d2);
    ck_assert(NULL!=newptr);
    free(newptr);
}
END_TEST

START_TEST(oomstub_realloc_upsize) {
    void* oldptr;
    void* newptr;
    int d1,d2;

    oldptr = malloc(1000);
    oomstub_setcountdown(0); /* No failure */
    d1 = oomstub_getcountdown();
    newptr = realloc(oldptr,2000);
    d2 = oomstub_getcountdown();
    oomstub_setcountdown(0);
    ck_assert(0==d1);
    ck_assert(0==d2);
    ck_assert(NULL!=newptr);
    free(newptr);

    oldptr = malloc(1000);
    oomstub_setcountdown(1); /* Immediate failure */
    d1 = oomstub_getcountdown();
    newptr = realloc(oldptr,2000);
    d2 = oomstub_getcountdown();
    oomstub_setcountdown(0);
    ck_assert(1==d1);
    ck_assert(0==d2);
    ck_assert(NULL==newptr);
    free(oldptr); /* cppcheck-suppress doubleFree */

    oldptr = malloc(1000);
    oomstub_setcountdown(2); /* Success first before failure */
    d1 = oomstub_getcountdown();
    newptr = realloc(oldptr,2000); /* cppcheck-suppress memleak */
    d2 = oomstub_getcountdown();
    oomstub_setcountdown(0);
    ck_assert(2==d1);
    ck_assert(1==d2);
    ck_assert(NULL!=newptr);
    free(newptr);
}
END_TEST

/************************************************************************/
/* Begining of test (potentially in child) */
void oomstub_checked_uninitialized_setup() {
    signals_catch();
    forktest_gprofdir();
}

/* End of test (potentially in child) */
void oomstub_checked_uninitialized_teardown() {
    signals_release();
}
/************************************************************************/
/* Begining of test case */
void oomstub_unchecked_setup() {
    forktest_init();
}
/* End of testcase */
void oomstub_unchecked_teardown() {
}
/************************************************************************/
Suite* oomstub_suite() {
    Suite *s;
    TCase *tc;

    s = suite_create("OOMStub");

    tc = tcase_create("Uninitialized");
    tcase_set_tags(tc,"oom");
    /* tcase_set_timeout(tc,5); */ /* seconds */
    tcase_add_checked_fixture(tc, oomstub_checked_uninitialized_setup, oomstub_checked_uninitialized_teardown);
    tcase_add_unchecked_fixture(tc, oomstub_unchecked_setup, oomstub_unchecked_teardown);
    suite_add_tcase(s, tc);

    tcase_add_test(tc, oomstub_malloc);
    tcase_add_test(tc, oomstub_calloc);
    tcase_add_test(tc, oomstub_realloc_malloc);
    tcase_add_test(tc, oomstub_realloc_free);
    tcase_add_test(tc, oomstub_realloc_downsize);
    tcase_add_test(tc, oomstub_realloc_upsize);
    return s;
}
/* vim: set tw=80: */
